@@ -26,7 +26,7 @@ flagReader.py options are still supported for backward compatibility
 """
 import cmd, re, os, sys, optparse, subprocess, tempfile, pyparsing, doctest, unittest, string
 from optparse import make_option
-__version__ = '0.4.4'
+__version__ = '0.4.5'
 
 class OptionParser(optparse.OptionParser):
     def exit(self, status=0, msg=None):
@@ -43,6 +43,15 @@ class OptionParser(optparse.OptionParser):
         """
         raise
         
+def remainingArgs(oldArgs, newArgList):
+    '''
+    >>> remainingArgs('-f bar   bar   cow', ['bar', 'cow'])
+    'bar   cow'
+    '''
+    pattern = '\s+'.join(newArgList) + '\s*$'
+    matchObj = re.search(pattern, oldArgs)
+    return oldArgs[matchObj.start():]
+
 def options(option_list):
     def option_setup(func):
         optionParser = OptionParser()
@@ -51,8 +60,8 @@ def options(option_list):
         optionParser.set_usage("%s [options] arg" % func.__name__.strip('do_'))
         def newFunc(instance, arg):
             try:
-                opts, newArgs = optionParser.parse_args(arg.split())
-                newArgs = (newArgs and arg[arg.find(newArgs[0]):]) or ''
+                opts, newArgList = optionParser.parse_args(arg.split())
+                newArgs = remainingArgs(arg, newArgList)
             except (optparse.OptionValueError, optparse.BadOptionError,
                     optparse.OptionError, optparse.AmbiguousOptionError,
                     optparse.OptionConflictError), e:
@@ -61,7 +70,13 @@ def options(option_list):
                 return
             if hasattr(opts, '_exit'):
                 return None
-            arg = arg.parser('%s %s%s%s' % (arg.parsed.command, newArgs, arg.parsed.terminator, arg.parsed.suffix))
+            terminator = arg.parsed.terminator
+            try:
+                if arg.parsed.terminator[0] == '\n':
+                    terminator = arg.parsed.terminator[0]
+            except IndexError:
+                pass
+            arg = arg.parser('%s %s%s%s' % (arg.parsed.command, newArgs, terminator, arg.parsed.suffix))
             result = func(instance, arg, opts)                            
             return result        
         newFunc.__doc__ = '%s\n%s' % (func.__doc__, optionParser.format_help())
@@ -232,7 +247,7 @@ class Cmd(cmd.Cmd):
     multilineCommands = []
     
     def _init_parser(self):
-        '''
+        r'''
         >>> c = Cmd()
         >>> c.multilineCommands = ['multiline']
         >>> c.caseInsensitive = True
@@ -372,12 +387,21 @@ class Cmd(cmd.Cmd):
           - multilineCommand: multiline
           - terminator: ;
         - terminator: ;
+        >>> print c.parser.parseString('multiline command ends\n\n').dump()
+        ['multiline', 'command ends', '\n', '\n']
+        - args: command ends
+        - multilineCommand: multiline
+        - statement: ['multiline', 'command ends', '\n', '\n']
+          - args: command ends
+          - multilineCommand: multiline
+          - terminator: ['\n', '\n']
+        - terminator: ['\n', '\n']
         '''
         outputParser = pyparsing.oneOf(['>>','>'])('output')
         terminatorParser = pyparsing.Or([(hasattr(t, 'parseString') and t) or pyparsing.Literal(t) for t in self.terminators])('terminator')
         stringEnd = pyparsing.stringEnd ^ '\nEOF'
-        multilineCommand = pyparsing.Or([pyparsing.Keyword(c, caseless=self.caseInsensitive) for c in self.multilineCommands])('multilineCommand')
-        oneLineCommand = pyparsing.Word(self.legalChars)('command')
+        self.multilineCommand = pyparsing.Or([pyparsing.Keyword(c, caseless=self.caseInsensitive) for c in self.multilineCommands])('multilineCommand')
+        oneLineCommand = (~self.multilineCommand + pyparsing.Word(self.legalChars))('command')
         pipe = pyparsing.Keyword('|', identChars='|')
         self.commentGrammars.ignore(pyparsing.sglQuotedString).ignore(pyparsing.dblQuotedString).setParseAction(lambda x: '')
         self.commentInProgress.ignore(pyparsing.sglQuotedString).ignore(pyparsing.dblQuotedString).ignore(pyparsing.cStyleComment)       
@@ -385,23 +409,27 @@ class Cmd(cmd.Cmd):
             pyparsing.Optional(pipe + pyparsing.SkipTo(outputParser ^ stringEnd)('pipeTo')) + \
             pyparsing.Optional(outputParser + pyparsing.SkipTo(stringEnd).setParseAction(lambda x: x[0].strip())('outputTo'))
         if self.caseInsensitive:
-            multilineCommand.setParseAction(lambda x: x[0].lower())
+            self.multilineCommand.setParseAction(lambda x: x[0].lower())
             oneLineCommand.setParseAction(lambda x: x[0].lower())
         if self.blankLinesAllowed:
-            blankLineTerminationParser = pyparsing.NoMatch
+            self.blankLineTerminationParser = pyparsing.NoMatch
         else:
-            blankLineTerminator = (pyparsing.Literal('\n')('terminator') + stringEnd)
-            blankLineTerminationParser = ((multilineCommand ^ oneLineCommand) + pyparsing.SkipTo(blankLineTerminator).setParseAction(lambda x: x[0].strip())('args') + blankLineTerminator)('statement')
-        multilineParser = (((multilineCommand ^ oneLineCommand) + pyparsing.SkipTo(terminatorParser).setParseAction(lambda x: x[0].strip())('args') + terminatorParser)('statement') +
-             pyparsing.SkipTo(outputParser ^ pipe ^ stringEnd).setParseAction(lambda x: x[0].strip())('suffix') + afterElements)
-        singleLineParser = ((oneLineCommand + pyparsing.SkipTo(terminatorParser ^ stringEnd ^ pipe ^ outputParser).setParseAction(lambda x:x[0].strip())('args'))('statement') +
-            pyparsing.Optional(terminatorParser) + afterElements)
+            self.blankLineTerminator = (pyparsing.lineEnd + pyparsing.lineEnd)('terminator')
+            self.blankLineTerminator.setResultsName('terminator')
+            self.blankLineTerminationParser = ((self.multilineCommand ^ oneLineCommand) + pyparsing.SkipTo(self.blankLineTerminator).setParseAction(lambda x: x[0].strip())('args') + self.blankLineTerminator)('statement')
+        self.multilineParser = (((self.multilineCommand ^ oneLineCommand) + pyparsing.SkipTo(terminatorParser).setParseAction(lambda x: x[0].strip())('args') + terminatorParser)('statement') +
+                                pyparsing.SkipTo(outputParser ^ pipe ^ stringEnd).setParseAction(lambda x: x[0].strip())('suffix') + afterElements)
+        self.singleLineParser = ((oneLineCommand + pyparsing.SkipTo(terminatorParser ^ stringEnd ^ pipe ^ outputParser).setParseAction(lambda x:x[0].strip())('args'))('statement') +
+                                 pyparsing.Optional(terminatorParser) + afterElements)
+        #self.multilineParser = self.multilineParser.setResultsName('multilineParser')
+        #self.singleLineParser = self.singleLineParser.setResultsName('singleLineParser')
+        #self.blankLineTerminationParser = self.blankLineTerminationParser.setResultsName('blankLineTerminatorParser')
         self.parser = (
             stringEnd |
-            multilineParser |
-            multilineCommand + pyparsing.SkipTo(stringEnd) |
-            singleLineParser |
-            blankLineTerminationParser
+            self.multilineParser |
+            self.singleLineParser |
+            self.blankLineTerminationParser | 
+            self.multilineCommand + pyparsing.SkipTo(stringEnd)
             )
         self.parser.ignore(pyparsing.sglQuotedString).ignore(pyparsing.dblQuotedString).ignore(self.commentGrammars).ignore(self.commentInProgress)
         
@@ -448,7 +476,7 @@ class Cmd(cmd.Cmd):
         if not line:
             return self.emptyline()
         if not pyparsing.Or(self.commentGrammars).setParseAction(lambda x: '').transformString(line):
-            return 0
+            return 0    # command was empty except for comments
         try:
             statement = self.parsed(line)
             while statement.parsed.multilineCommand and (statement.parsed.terminator == ''):
@@ -615,7 +643,12 @@ class Cmd(cmd.Cmd):
             self.stdout.write('%s - was: %s\nnow: %s\n' % (paramName, currentVal, val))
         except (ValueError, AttributeError, NotSettableError), e:
             self.do_show(arg)
+    do_set.__doc__ = '%s\nOne of: %s' % (do_set.__doc__, ', '.join(settable))
                 
+    def do_pause(self, arg):
+        'Displays the specified text then waits for the user to press RETURN.'
+        raw_input(arg + '\n')
+        
     def do_shell(self, arg):
         'execute a command as if at the OS prompt.'
         os.system(arg)
@@ -886,37 +919,51 @@ class Cmd2TestCase(unittest.TestCase):
             self.cmdapp = self.CmdApp()
             try:
                 tfile = open(self.transcriptFileName)
-                self.transcript = tfile.readlines()
+                self.transcript = iter(tfile.readlines())
                 tfile.close()
             except IOError:
                 self.transcript = []
-    def divideTranscript(self):
-        self.dialogue = []
-        commandStart = None
-        response = []
-        command = []
-        prompt = self.cmdapp.prompt.rstrip()
-        continuationPrompt = self.cmdapp.continuationPrompt.rstrip()
-        for (lineNum, line) in enumerate(self.transcript):
-            if line.startswith(prompt):
-                if command:
-                    self.dialogue.append((commandStart, ''.join(command), ''.join(response)))
-                command = [line[len(prompt)+1:]]
-                commandStart = lineNum
-                response = []
-            elif line.startswith(continuationPrompt):
-                command.append(line[len(continuationPrompt)+1:])
-            else:
-                response.append(line)
+    def assertEqualEnough(self, got, expected, message):
+        got = got.strip().splitlines()
+        expected = expected.strip().splitlines()
+        self.assertEqual(len(got), len(expected), message)
+        for (linegot, lineexpected) in zip(got, expected):
+            matchme = re.escape(lineexpected.strip()).replace('\\*', '.*').  \
+                      replace('\\ ', ' ')
+            self.assert_(re.match(matchme, linegot.strip()), message)
     def testall(self):
-        if self.CmdApp:            
-            self.divideTranscript()        
-            for (lineNum, command, expected) in self.dialogue:
-                self.cmdapp.onecmd(command)
-                result = self.outputTrap.read()
-                self.assertEqual(result.strip(), expected.strip(), 
-                    '\nFile %s, line %d\nCommand was:\n%s\nExpected:\n%s\nGot:\n%s\n' % 
-                    (self.transcriptFileName, lineNum, command, expected, result))    
+        if self.CmdApp:
+            lineNum = 0
+            try:
+                line = self.transcript.next()
+                while True:
+                    while not line.startswith(self.cmdapp.prompt):
+                        line = self.transcript.next()
+                    command = [line[len(self.cmdapp.prompt):]]
+                    line = self.transcript.next()
+                    while line.startswith(self.cmdapp.continuationPrompt):
+                        command.append(line[len(self.cmdapp.continuationPrompt):])
+                        line = self.transcript.next()
+                    command = ''.join(command)
+                    self.cmdapp.onecmd(command)
+                    result = self.outputTrap.read()
+                    if line.startswith(self.cmdapp.prompt):
+                        self.assertEqualEnough(result.strip(), '', 
+                            '\nFile %s, line %d\nCommand was:\n%s\nExpected: (nothing) \nGot:\n%s\n' % 
+                            (self.transcriptFileName, lineNum, command, result))    
+                        continue
+                    expected = []
+                    while not line.startswith(self.cmdapp.prompt):
+                        expected.append(line)
+                        line = self.transcript.next()
+                    expected = ''.join(expected)
+                    self.assertEqualEnough(expected.strip(), result.strip(), 
+                        '\nFile %s, line %d\nCommand was:\n%s\nExpected:\n%s\nGot:\n%s\n' % 
+                        (self.transcriptFileName, lineNum, command, expected, result))    
+                    # this needs to account for a line-by-line strip()ping
+            except StopIteration:
+                pass
+                # catch the final output?
     def tearDown(self):
         if self.CmdApp:
             self.outputTrap.tearDown()
